import {
    EVENT_SERVICE,
} from "./../../providers/service.providers";
import { EventService } from "../../services/event.service";
import {
    Component,
    OnInit,
    Inject,
    OnDestroy
} from "@angular/core";
import { ActivatedRoute } from "@angular/router";
import { Event } from "../../models/Event";
import { ImageHelper } from "../../helpers/ImageHelper";
import { environment } from "./../../../environments/environment";
import { LabelsService } from "src/app/services/labels.service";
import { SessionCartService } from "./session-cart/session-cart.service";
import { Session } from "src/app/models/Session";
import { ActiveEventService } from "src/app/services/active-event.service";
import { LocalizableError } from 'src/app/models/LocalizableError';

@Component({
    selector: "app-event",
    templateUrl: "./event.component.html",
    styleUrls: ["./event.component.css"]
})
export class EventComponent implements OnInit {
    public EventDetailSections = EventDetailSections;
    public readableEventId: string;
    public event: Event;
    public sessions: Session[];
    public selectedSection: EventDetailSections = EventDetailSections.Sessions;
    public isSessionCartCollapsed: Boolean = false;
    public isSessionCartAllowed: Boolean = false;
    public isLoadingData = false;
    public error: LocalizableError;

    private defaultImageUrlSelfHosted = '/assets/images/default-event-image.jpg';
    private defaultImageUrlCrmHosted = 'homehero.jpg';

    constructor(
        private route: ActivatedRoute,
        @Inject(EVENT_SERVICE) private eventService: EventService,
        private imageHelper: ImageHelper,
        private labelsService: LabelsService,
        private sessionCartService: SessionCartService,
        private activeEventService: ActiveEventService
    ) {}

    ngOnInit() {
        this.route.queryParamMap.subscribe(paramMap => {
            this.readableEventId = paramMap.get(
                EventQueryParameterNames.ReadableEventId
            );

            this.getEventAndSessionData().then(() =>
                this.setIsSessionCartAllowed()
            );
        });

        this.sessionCartService.isOpen = false;
    }

    private handleErrorResponse(error: LocalizableError) {
        this.isLoadingData = false;
        this.error = error;
    }

    public getBannerImage() {
        if (this.event.image != null) {
            return this.event.image;
        } else {
            if (environment.useRestStack === true) {
                return this.defaultImageUrlSelfHosted;
            } else {
                return this.imageHelper.getImage(
                    null,
                    this.defaultImageUrlCrmHosted
                );
            }
        }
    }

    public registerForAllSessions() {
        this.sessions.forEach(session => {
            this.sessionCartService.addSession(this.readableEventId, session);
        });

        this.sessionCartService.isOpen = true;
    }

    private async getEventAndSessionData(): Promise<void> {
        try {
            this.event = await this.activeEventService.getEvent(this.readableEventId).toPromise();

            this.sessions = await this.activeEventService.getSessions(this.readableEventId).toPromise();
        } catch (error) {
            return console.error(error);
        }
    }

    private async setIsSessionCartAllowed(): Promise<void> {
        const isFreeEvent = await this.activeEventService.isFreeEvent(this.readableEventId);
        const hasSessions = await this.activeEventService.hasEventSessions(this.readableEventId);

        this.isSessionCartAllowed = this.sessionCartService.isSessionCartEnabledForEvent(
            this.event,
            isFreeEvent,
            hasSessions
        );
    }
}


export enum EventQueryParameterNames {
    ReadableEventId = "id",
    SpeakerId = "speakerId"
}

export enum EventDetailSections {
    Sessions = "sessions",
    SessionTracks = "session-tracks",
    Speakers = "speakers",
    PassInformation = "pass-information"
}
