import { Injectable, Output, EventEmitter } from "@angular/core";
import { Session } from "src/app/models/Session";
import { Event } from "src/app/models/Event";

@Injectable()
export class SessionCartService {
    @Output() isOpenChange: EventEmitter<boolean> = new EventEmitter();
    @Output() sessionCartChange: EventEmitter<Array<Session>> = new EventEmitter();
    @Output() cntActiveAttendeeRegistrationsChange: EventEmitter<number> = new EventEmitter();

    private _isOpen = false;
    private sessionCarts: Map<string, Array<Session>> = new Map();
    private cntActiveAttendeeRegistrations: Map<string, number> = new Map();

    private readonly CART_KEY_PREFIX = "cart-";

    public get isOpen(): boolean {
        return this._isOpen;
    }

    public set isOpen(isOpen: boolean) {
        this._isOpen = isOpen;
        this.isOpenChange.emit(isOpen);
    }

    public getCntActiveAttendeeRegistrations(readableEventId: string): number {
        if (this.cntActiveAttendeeRegistrations.has(readableEventId)) {
            return this.cntActiveAttendeeRegistrations.get(readableEventId);
        }

        return 0;
    }

    public setCntActiveAttendeeRegistrations(readableEventId: string, count: number) {
        this.cntActiveAttendeeRegistrations.set(readableEventId, count);
        this.cntActiveAttendeeRegistrationsChange.emit(count);
    }

    public isSessionCartEnabledForEvent(event: Event, isFreeEvent: boolean, hasSessions: boolean): boolean {
        return event.allowCustomAgenda && isFreeEvent && hasSessions && !event.showWaitlist;
    }

    public getSessionCartForEvent(readableEventId: string): Array<Session> {
        this.loadCartFromLocalStorage(readableEventId);

        if (!this.sessionCarts.has(readableEventId)) {
            this.sessionCarts.set(readableEventId, new Array<Session>());
        }

        return this.sessionCarts.get(readableEventId);
    }

    public addSession(readableEventId: string, session: Session) {
        if (!this.sessionCarts.has(readableEventId)) {
            this.sessionCarts.set(readableEventId, new Array<Session>());
        }

        if (!this.isSessionInCart(readableEventId, session)) {
            this.sessionCarts.get(readableEventId).push(session);
            this.sessionCartChange.emit(Array.from(this.sessionCarts.get(readableEventId).values()));
        }

        this.saveCartToLocalStorage(readableEventId);
    }

    public removeSession(readableEventId: string, session: Session) {
        if (!this.sessionCarts.has(readableEventId)) {
            this.sessionCarts.set(readableEventId, new Array<Session>());
        }

        if (this.isSessionInCart(readableEventId, session)) {
            const index = this.sessionCarts.get(readableEventId).findIndex(s => s.id === session.id);
            this.sessionCarts.get(readableEventId).splice(index, 1);

            this.sessionCartChange.emit(Array.from(this.sessionCarts.get(readableEventId).values()));
        }

        this.saveCartToLocalStorage(readableEventId);
    }

    private isSessionInCart(readableEventId, session: Session) {
        if (this.sessionCarts.has(readableEventId)) {
            return this.sessionCarts.get(readableEventId).findIndex(s => s.id === session.id) >= 0;
        }

        return false;
    }

    public flushCart(readableEventId: string) {
        this.sessionCarts.delete(readableEventId);
        localStorage.removeItem(this.getLocalStorageKeyForEvent(readableEventId));
    }

    private saveCartToLocalStorage(readableEventId: string) {
        if (this.sessionCarts.has(readableEventId)) {
            localStorage.setItem(
                this.getLocalStorageKeyForEvent(readableEventId),
                JSON.stringify(this.sessionCarts.get(readableEventId))
            );
        }
    }

    private loadCartFromLocalStorage(readableEventId: string) {
        const isCartForEventInStorage = localStorage.hasOwnProperty(this.getLocalStorageKeyForEvent(readableEventId));

        if (isCartForEventInStorage) {
            const storageCart = localStorage.getItem(this.getLocalStorageKeyForEvent(readableEventId));
            this.sessionCarts.set(readableEventId, JSON.parse(storageCart));

            this.sessionCartChange.emit(this.sessionCarts.get(readableEventId));
        }
    }

    private getLocalStorageKeyForEvent(readableEventId: string) {
        return this.CART_KEY_PREFIX + readableEventId;
    }
}
