import { Component, OnInit, Inject, ViewChild } from '@angular/core';
import { ActivatedRoute, Router } from '@angular/router';
import { EVENT_SERVICE, USER_SERVICE } from './../../providers/service.providers';
import { CaptchaService } from './../../services/captcha.service';
import { Attendee } from '../../models/Attendee';
import { Event } from '../../models/Event';
import { Pass } from '../../models/Pass';
import { RegistrationData } from '../../models/RegistrationData';
import { HipObject } from '../../models/HipObject';
import * as CustomRegistrationFieldModel from '../../models/CustomRegistrationField';
import { EventService } from '../../services/event.service';
import { LabelsService } from '../../services/labels.service';
import { BrowserSessionService } from '../../services/browser-session.service';
import { AttendeeComponent } from './attendee/attendee.component';
import { UserService } from 'src/app/services/user.service';
import { ExternalLoginUrlResolver } from 'src/app/resolvers/external.login.url.resolver';
import { Session } from 'src/app/models/Session';
import { ActiveEventService } from 'src/app/services/active-event.service';
import { SessionCartService } from '../event/session-cart/session-cart.service';
import { TranslationKeys } from 'src/app/constants/TranslationKeys';
import { LocalizableError } from 'src/app/models/LocalizableError';

@Component({
    selector: 'app-eventregistration',
    templateUrl: './eventregistration.component.html',
    styleUrls: ['./eventregistration.component.css']
})
export class EventRegistrationComponent implements OnInit {
    @ViewChild(AttendeeComponent) attendeeComponent: AttendeeComponent;

    public cntActiveDataRetrievals = 0;
    public error: LocalizableError;

    attendees: Attendee[];
    attendeeEditingCount: number;
    waitlistedAttendees: Attendee[];
    event: Event;
    passes: Pass[];
    registrationCount: number;
    readableEventId: string;
    registrationInProgress: boolean;
    customRegistrationFields: CustomRegistrationFieldModel.CustomRegistrationField[];
    isJapanese: boolean;

    private readonly formValidStateString = 'VALID';
    private total = 0.0;
    private currencySymbol = '$';
    private formValidationStatus: string;
    private isUserLoggedIn: boolean;

    constructor(
        private activeEventService: ActiveEventService,
        private sessionCartService: SessionCartService,
        @Inject(EVENT_SERVICE) private eventService: EventService,
        private captchaService: CaptchaService,
        private route: ActivatedRoute,
        private router: Router,
        private labelsService: LabelsService,
        private browserSessionService: BrowserSessionService,
        @Inject(USER_SERVICE) private userService: UserService
    ) {
        this.attendees = [];
        this.waitlistedAttendees = [];
        this.attendeeEditingCount = 0;
    }

    ngOnInit() {
        this.readableEventId = this.route.snapshot.queryParams['id'];
        this.loadEvent();
        this.loadCustomRegistrationFields();

        this.labelsService.getLabelsModel().subscribe(labelsModel => {
            this.isJapanese = labelsModel.isJapanese;
        });

        this.userService.isLoggedIn().subscribe(isUserLoggedIn => {
            this.isUserLoggedIn = isUserLoggedIn;
        });

        this.sessionCartService.setCntActiveAttendeeRegistrations(this.readableEventId, 0);
    }

    public isLoginProcessActive() {
        return ExternalLoginUrlResolver.isResolverActive();
    }

    public mustUserSignIn() {
        return !(this.event.allowAnonymousRegistrations || this.isUserLoggedIn);
    }

    public clearForms(): void {
        let selectorResults = document.querySelectorAll('.event-customregistrationsfields-container input');
        let selectorResult: any;
        let i: number;
        for (i = 0; i < selectorResults.length; i++) {
            selectorResult = selectorResults[i];
            selectorResult.value = '';
            selectorResult.dispatchEvent(new Event('input'));
        }

        selectorResults = document.querySelectorAll('.event-customregistrationsfields-container input:checked');
        for (i = 0; i < selectorResults.length; i++) {
            selectorResult = selectorResults[i];
            selectorResult.checked = false;
        }

        selectorResults = document.querySelectorAll('.event-customregistrationsfields-container select');
        for (i = 0; i < selectorResults.length; i++) {
            selectorResult = selectorResults[i];
            selectorResult.selectedIndex = 0;
        }

        this.sessionCartService.setCntActiveAttendeeRegistrations(this.readableEventId, 0);
    }

    /**
     * Checks if the validation status of the attendee form (child view) is valid.
     */
    public isAttendeeFormValid(): boolean {
        return this.formValidationStatus === this.formValidStateString;
    }

    /**
     * Checks if the checkout button should be disabled.
     */
    public isCheckoutButtonDisabled(): boolean {
        // NOTE: order of if statements is very important for logic.
        if (this.registrationInProgress) {
            return true;
        }

        if (this.isAttendeeCountExceedingEventCapacity()) {
            return true;
        }

        if (this.isAttendeeFormValid() && this.event.showWaitlist && this.hasAttendeeCountReachedEventCapacity()) {
            // Case: Event Full, 0 Event Attendees, 1 Waitlist attendee (did not press 'add another attendee')
            return false;
        }

        if (this.isAttendeeFormValid() && !this.hasAttendeeCountReachedEventCapacity()) {
            return false;
        }

        if (this.attendees.length === 0 && this.waitlistedAttendees.length === 0) {
            return true;
        }

        return false;
    }

    private loadEvent(): void {
        this.cntActiveDataRetrievals += 1;
        this.activeEventService.getEvent(this.readableEventId).subscribe(event => {
                this.event = event;

                if (this.event) {
                    this.loadEventRegistrationCount();
                    this.loadEventPasses();
                }

                this.cntActiveDataRetrievals -= 1;
            },
            (error: LocalizableError) => this.handleDataRetrievalError(error)
        );
    }

    private loadCustomRegistrationFields() {
        this.cntActiveDataRetrievals += 1;

        this.eventService.getCustomRegistrationFields(this.readableEventId).subscribe(
            customRegistrationFields => {
                this.customRegistrationFields = [];
                for (const customRegistrationField of customRegistrationFields) {
                    this.customRegistrationFields.push(
                        Object.assign(
                            new CustomRegistrationFieldModel.CustomRegistrationField(),
                            customRegistrationField
                        )
                    );
                }
                this.cntActiveDataRetrievals -= 1;
            },
            (error: LocalizableError) => this.handleDataRetrievalError(error)
        );
    }

    private loadEventPasses(): void {
        this.activeEventService.getPasses(this.readableEventId).subscribe(
            passes => {
                this.passes = passes;
                for (const pass of this.passes) {
                    pass.passesUsed = 0;
                }
            },
            (error: LocalizableError) => this.handleDataRetrievalError(error));
    }

    private loadEventRegistrationCount(): void {
        this.cntActiveDataRetrievals += 1;

        this.eventService.getEventRegistrationCount(this.readableEventId).subscribe(
            registrationCount => {
                this.registrationCount = registrationCount;

                if (this.hasAttendeeCountReachedEventCapacity() && !this.event.showWaitlist) {
                    this.showEventCapacityReachedErrorMessage();
                }

                if (this.isAttendeeCountExceedingEventCapacity()) {
                    this.showEventCapacityExceedingErrorMessage();
                }

                this.cntActiveDataRetrievals -= 1;
            },
            (error: LocalizableError) => this.handleDataRetrievalError(error)
        );
    }

    private handleDataRetrievalError(error: LocalizableError) {
        if (this.cntActiveDataRetrievals > 0) {
            this.cntActiveDataRetrievals -= 1;
        }

        this.handleErrorResponse(error);
    }

    private handleErrorResponse(error: LocalizableError) {
        this.error = error;
    }

    public showWaitlist(): boolean {
        if (!this.paidEvent()) {
            return (
                this.event.showWaitlist &&
                this.event.isCapacityRestricted &&
                (this.hasAttendeeCountReachedEventCapacity() ||
                    this.waitlistedAttendees.length > 0)
            );
        }

        return this.event.showWaitlist && (this.allPassesSoldOut() || this.hasAttendeeCountReachedEventCapacity());
    }

    public isAttendeeCountExceedingEventCapacity(): boolean {
        return this.getAvailableEventCapacity() < 0;
    }

    public hasAttendeeCountReachedEventCapacity(): boolean {
        return this.getAvailableEventCapacity() === 0;
    }

    /**
     * Returns the available capacity of the event. Attendees in checkout process are included.
     * If max capacity is not restricted, MAX_INT is returned.
     */
    private getAvailableEventCapacity(): number {
        if (!this.event.isCapacityRestricted) {
            return Number.MAX_SAFE_INTEGER;
        }

        return this.event.maxCapacity - this.registrationCount - this.attendees.length;
    }

    private allPassesSoldOut(): boolean {
        for (const pass of this.passes) {
            if (pass.numberOfPassesLeft > pass.passesUsed) {
                return false;
            }
        }

        return true;
    }

    public addAttendee(attendee: Attendee, waitlisted: boolean): void {
        if (this.isAttendeeCountExceedingEventCapacity()) {
            this.showEventCapacityExceedingErrorMessage();
            return;
        }

        if (waitlisted) {
            this.waitlistedAttendees.push(attendee);
        } else {
            this.attendees.push(attendee);
            this.clearForms();
        }

        if (this.isAttendeeCountExceedingEventCapacity()) {
            this.showEventCapacityExceedingErrorMessage();
        }

        if (this.hasAttendeeCountReachedEventCapacity()) {
            this.showEventCapacityReachedErrorMessage();
        }

        this.updateTotal(null, attendee.passId);
        this.updateActiveAttendeeRegistrationsCount();
    }

    private removeAttendee(attendee: Attendee, waitlisted: boolean): void {
        const index = this.findAttendeeIndex(attendee, waitlisted);
        if (index !== -1) {
            if (waitlisted) {
                this.waitlistedAttendees.splice(index, 1);
            } else {
                const removedPassId = this.attendees[index].passId;
                this.attendees.splice(index, 1);

                // Move the first waitlisted attendee to the registrants
                if (this.waitlistedAttendees.length > 0) {
                    if (this.paidEvent()) {
                        /* for paid events we cannot just move the first waitlisted attendee to attendees
                      as there may be a conflict with passes try to find a waitlisted attendee that wants
                      to buy the same pass that the removed attendee had and move that one */
                        const idx = this.indexOfFirstWaitlistedAttendeeWithPassId(removedPassId);
                        if (idx !== -1) {
                            this.attendees.push(this.waitlistedAttendees[idx]);
                            this.waitlistedAttendees.splice(idx, 1);
                        }
                    } else {
                        // for unpaid events we can just move any of the waitlisted attendees
                        this.attendees.push(this.waitlistedAttendees[0]);
                        this.waitlistedAttendees.splice(0, 1);
                    }
                }
            }

            this.updateTotal(attendee.passId, null);
        }

        this.clearForms();
        this.updateActiveAttendeeRegistrationsCount();
    }

    private updateActiveAttendeeRegistrationsCount() {
        const total = this.attendees.length + this.waitlistedAttendees.length;
        this.sessionCartService.setCntActiveAttendeeRegistrations(this.readableEventId, total);
    }

    /**
     * Finds the index of the first waitlisted attendee with a specified pass
     * @param passId The pass id
     */
    private indexOfFirstWaitlistedAttendeeWithPassId(passId: string) {
        return this.waitlistedAttendees.findIndex(a => a.passId === passId);
    }

    /**
     * Event handler for the update attendee event of the attendee component
     * @param attendees An array of length 2, the first item being the attendee before the update operation,
     * the second is the attendee after the update
     * @param waitlisted Whether the attendee updated is in the waitlist or not
     */
    private updateAttendee(attendees: Attendee[], waitlisted: boolean): void {
        const index = this.findAttendeeIndex(attendees[0], waitlisted);
        if (index !== -1) {
            if (waitlisted) {
                this.waitlistedAttendees[index] = attendees[1];
            } else {
                const attendee = attendees[1];
                this.attendees[index] = attendee;
                this.clearForms();
            }

            this.updateTotal(attendees[0].passId, attendees[1].passId);
        }
    }

    private attendeeEditingStateChanged(newEditingState: boolean) {
        if (newEditingState) {
            this.attendeeEditingCount++;
        } else {
            if (this.attendeeEditingCount > 0) {
                this.attendeeEditingCount--;
            }
        }
    }

    /**
     * Finds the index of the attendee
     * @param attendee The attendee
     * @param waitlisted If true the waitlisted attendee list will be searched, otherwise the attendee list
     */
    private findAttendeeIndex(attendee: Attendee, waitlisted: boolean): number {
        return waitlisted
            ? this.waitlistedAttendees.findIndex(
                  a =>
                      a.firstName === attendee.firstName &&
                      a.lastName === attendee.lastName &&
                      a.email === attendee.email
              )
            : this.attendees.findIndex(
                  a =>
                      a.firstName === attendee.firstName &&
                      a.lastName === attendee.lastName &&
                      a.email === attendee.email
              );
    }

    private updateTotal(passIdToRemove: string, passIdToAdd: string): void {
        if (passIdToRemove != null) {
            this.total -= this.findPassValue(passIdToRemove);
        }

        if (passIdToAdd != null) {
            this.total += this.findPassValue(passIdToAdd);
        }
    }

    private findPassValue(passId: string): number {
        if (this.passes == null) {
            return 0.0;
        }
        const pass: Pass = this.passes.find(p => p.passId === passId);
        if (pass) {
            this.currencySymbol = pass.currencySymbol;
            return pass.price;
        } else {
            return 0.0;
        }
    }

    public checkout(): void {
        this.registrationInProgress = true;

        // reload event registration count in case there happened registrations in the meantime.
        this.eventService.getEventRegistrationCount(this.readableEventId).subscribe(
            eventRegistrationCount => {
                this.registrationCount = eventRegistrationCount;

                if (this.isAttendeeFormValid()) {
                    this.attendeeComponent.confirmEdit();
                }

                if (this.isAttendeeCountExceedingEventCapacity()) {
                    this.showEventCapacityExceedingErrorMessage();
                    this.registrationInProgress = false;
                    return;
                }

                const registrationResult = this.createRegistrationData();
                 this.registerToEvent(registrationResult);
            },
            (error: LocalizableError) => {
                this.registrationInProgress = false;
                this.handleDataRetrievalError(error)
            }
        );
    }

    private createRegistrationData(): RegistrationData {
        let hipObjectResult: HipObject = CaptchaService.EmptyHipObject;

        if (this.event.enableCaptcha) {
            hipObjectResult = this.captchaService.getHipObject();
        }

        if (this.paidEvent()) {
            this.browserSessionService.setCaptcha(hipObjectResult);
        }

        if (this.isAttendeeFormValid()) {
            this.attendeeComponent.confirmEdit();
        }

        let attendeesToSend: Attendee[] = [];
        attendeesToSend = attendeesToSend.concat(this.attendees);
        attendeesToSend = attendeesToSend.concat(this.waitlistedAttendees);

        const sessions = this.sessionCartService.getSessionCartForEvent(this.readableEventId);
        const isSessionCartEnabled = this.sessionCartService.isSessionCartEnabledForEvent(this.event, !this.paidEvent(), sessions.length > 0)

        if (isSessionCartEnabled && sessions) {
            attendeesToSend.forEach(attendee => {
                    attendee.sessions = sessions.map(s => s.id);
            });
        }

        const registrationData: RegistrationData = {
            attendees: attendeesToSend,
            hipObject: hipObjectResult
        };

        if (!this.event.enableCaptcha) {
            registrationData.hipObject = undefined;
        }

        return registrationData;
    }

    private registerToEvent(registrationData: RegistrationData) {
        this.eventService
        .registerToEvent(this.readableEventId, registrationData)
        .subscribe(
            registrationResult => {
                if (registrationResult.status === 'Success') {
                    this.router.navigate(['../confirmation'], {
                        relativeTo: this.route,
                        queryParams: {
                            id: this.event.readableEventId
                        }
                    });
                } else if (registrationResult.status === 'Initiated') {
                    this.browserSessionService.setPurchaseId(registrationResult.purchaseId);
                    this.browserSessionService.setRegistrationTotal(this.total);
                    this.browserSessionService.setCurrency(this.currencySymbol);
                    this.router.navigate(['../payment'], {
                        relativeTo: this.route,
                        queryParams: {
                            id: this.event.readableEventId,
                            currencySymbol: this.currencySymbol,
                            purchaseId: registrationResult.purchaseId
                        }
                    });
                } else if (registrationResult.status === 'Redirect') {
                    this.browserSessionService.setPurchaseId(registrationResult.purchaseId);
                    this.browserSessionService.setRegistrationTotal(this.total);
                    this.browserSessionService.setCurrency(this.currencySymbol);
                    window.location.href = registrationResult.redirectUrl;
                } else {
                    let errorMessage = "Event registration failed.";
                    if (registrationResult.errorMessage != null && registrationResult.errorMessage != "") {
                        errorMessage = registrationResult.errorMessage;
                    }

                    this.error = new LocalizableError(errorMessage, TranslationKeys.RegistrationFailed);
                    this.registrationInProgress = false;
                }
            },
            (error: LocalizableError) => this.handleErrorResponse(error)
        )
        .add(() => {
            this.registrationInProgress = false;
        });
    }

    private showEventCapacityExceedingErrorMessage() {
        this.error = new LocalizableError(
            'You cannot register that many attendees because the maximum event capacity is already reached.',
            TranslationKeys.EventCapacityExceeding
        );
    }

    private showEventCapacityReachedErrorMessage() {
        this.error = new LocalizableError(
            'The events maximum capacity has been reached.',
            TranslationKeys.EventCapacityReached
        );
    }

    private paidEvent(): boolean {
        if (this.passes == null) {
            return false;
        }

        return this.passes.length > 0;
    }

    private autoregisterWaitlistItems(): boolean {
        return this.event.autoregisterWaitlistItems;
    }

    private showAutomaticRegistrationCheckbox(): boolean {
        return this.event.showAutomaticRegistrationCheckbox;
    }

    private changeFormValidationStatus(validationStatus) {
        this.formValidationStatus = validationStatus;
    }
}

export enum EventCheckoutRoutePaths {
    REGISTRATION = "event/registration",
    CONFIRMATION = "event/confirmation",
    PAYMENT = "event/payment",
    SUCCESS_PAYMENT = "event/successpayment"
}
